/********************************************************************
 * File: KDTree.h
 * Author: TODO: Your name here!
 *
 * An interface representing a kd-tree in some number of dimensions.
 * The tree can be constructed from a set of data and then queried
 * for membership and nearest neighbors.
 */

#ifndef KDTree_Included
#define KDTree_Included

#include <vector>
#include <string>
#include <deque>
#include <iterator>
#include "BoundedPQueue.h"
using namespace std;

class KDTree
{
public:
    /* Type: value_type
     * ----------------------------------------------------
     * The type of the elements stored in this KDTree.  This
     * is a pair of a vector<double> (key) and a string (value).
     * You will see this type extensively in the course of this
     * assignment.
     */
    typedef pair<vector<double>, string> value_type;

    /* Constructor: KDTree(InputIterator, InputIterator)
     * Usage: KDTree myTree(data.begin(), data.end())
     * ----------------------------------------------------
     * Constructs a KDTree from a range of point/value pairs.
     * Each entry in the range must be a value_type, but can
     * come from any source.
     */
    template <typename InputIterator>
        KDTree(InputIterator begin, InputIterator end);

    /* Destructor: ~KDTree()
     * Usage: (implicit)
     * ----------------------------------------------------
     * Cleans up all resources used by the KDTree.  Note that
     * if you implement copy behavior using reference-counting,
     * the destructor might not actually free any memory.
     */
    ~KDTree();

    /* Type: const_iterator
     * ----------------------------------------------------
     * An iterator that can traverse the elements of a kd-tree
     * but cannot write to them.
     */
    class const_iterator;

    /* const_iterator begin() const;
     * const_iterator end() const;
     * Usage: for(KDTree::const_iterator itr = kd.begin();
     *            itr != kd.end(); ++itr)
     * ----------------------------------------------------
     * Returns iterators spanning the range of the KDTree.
     * The elements are not stored in any particular order.
     */
    const_iterator begin() const;
    const_iterator end()   const;

    /* size_t dimension() const;
     * Usage: vector<double> entry(kd.dimension());
     * ----------------------------------------------------
     * Returns the dimension of the vectors stored in the
     * kd-tree.  For example, if the tree holds points in
     * two-dimensional space, this function would return
     * two.
     */
    size_t dimension() const;

    /* const_iterator find (const vector<double>& key) const;
     * Usage: if (kd.find(v) != kd.end())
     * ----------------------------------------------------
     * Returns an iterator to the key/value pair with the
     * specified key, or end() if no entry in the kd-tree
     * has the specified key.
     */
    const_iterator find(const vector<double>& key) const;

    /* size_t count(const vector<double>& key) const;
     * Usage: if (kd.count(v))
     * ----------------------------------------------------
     * Returns 1 if the specified point is in the tree and 0
     * otherwise.
     */
    size_t count(const vector<double>& key) const;

    /* size_t size()  const;
     * bool   empty() const;
     * Usage: if (kd.empty())
     * ----------------------------------------------------
     * Returns the number of elements in the kd-tree and
     * whether the tree is empty.
     */
    size_t size() const;
    bool   empty() const;

    /* string kNearestNeighborValue(const vector<double>& key, size_t k) const
     * Usage: cout << kd.kNearestNeighborValue(v, 3) << endl;
     * ----------------------------------------------------
     * Given a point v and an integer k, finds the k points
     * in the KDTree nearest to v and returns the most common
     * string associated with those points.  In the event of
     * a tie, one of the most frequent strings will be chosen.
     */
    string kNearestNeighborValue(const vector<double>& key, size_t k) const;

private:

    /* Type: KDTreeNode
     * ----------------------------------------------------
     * A node in the kd-tree.  The node stores three pieces
     * of information:
     * 
     * 1. The location and string associated with that location.
     * 2. The left and right subtrees.
     * 3. The parent node.
     *
     * The parent node pointer exists in order for the
     * const_iterator type to work correctly.  You will
     * probably not need to use it in the course of this
     * assignment.
     *
     * Note that although each node has two children, the node
     * itself does not store which index it splits on.  This
     * information will be communicated through the recursive
     * calls that traverse the tree.  However, irrespective of
     * which axis the node splits on, the left subtree contains
     * points whose component is _strictly less_ than the
     * compared component of the current node and the right subtree
     * contains points whose component is _greater than or equal
     * to_ the current node's compared component.
     */
    struct KDTreeNode
    {
        /* Constructor: KDTreeNode(const value_type&, KDTreeNode*);
         * Usage: KDTreeNode* node = new KDTreeNode(value, parentNode);
         * ------------------------------------------------
         * Constructs a new KDTreeNode that stores the indicated
         * value and has the specified parent.  Note that this
         * constructor does _not_ update any fields of the parent
         * node.
         */
        KDTreeNode(const value_type&, KDTreeNode* parent);

        /* Destructor: ~KDTreeNode();
         * Usage: (implicit)
         * ------------------------------------------------
         * Recursively deletes both child nodes.
         */
        ~KDTreeNode();

        /* The point and associated value stored at this point. */
        value_type value;

        /* Left subtree: Points whose nth component is strictly less
         * than this node's.
         */
        KDTreeNode* left;

        /* Right subtree: Points whose nth component is greater than
         * or equal to this node's.
         */
        KDTreeNode* right;

        /* Parent node.  This is primarily used by the const_iterator
         * type and you will most likely not use it in your solution.
         */
        KDTreeNode* parent;
    };

    KDTreeNode* root;          // The root of the tree

    size_t      numDimensions; // The number of dimensions of the vectors stored
                               // in this tree.

    size_t      numElems;      // The number of points stored in the tree.

    /* A comparison function which returns whether two pointers to value_types point
     * to value_types that have the same vector.  This is used in the constructor
     * to eliminate duplicate vectors.  You will probably not need to use this function
     * in your solution.
     */
    static bool ValuesHaveEqualPoints(const value_type* one, const value_type* two);

    /* A comparison function which returns whether the object pointed at by the first
     * pointer is less than the object pointed at by the second pointer.  As with the
     * previous function, you will probably not need to use this function in your
     * solution.
     */
    static bool CompareValueTypesIndirectly(const value_type* one, const value_type* two);

    /* Function: KDTreeNode* recMakeKDTree(vector<value_type*>::iterator begin,
     *                                     vector<value_type*>::iterator end,
     *                                     size_t axis, KDTreeNode* parent);
     * ----------------------------------------------------
     * This function is responsible for transforming a range of
     * data into a kd-tree.  The first two parameters delineate a
     * range of values to build into the KD tree.  The third parameter
     * describes which axis the next split should be made on.  The
     * final parameter is the node that will be used as the parent
     * of the recursively-generated tree.
     *
     * This function has already been written and will be given to you.
     * However, one piece of the function (the AxisCompare functor) has
     * only a skeletal implementation.  Your first task will be to implement
     * this functor.
     */
    KDTreeNode* recMakeKDTree(vector<value_type*>::iterator begin,
                              vector<value_type*>::iterator end,
                              size_t axis, KDTreeNode* parent);
};

/* Type: KDTree::const_iterator
 * --------------------------------------------------------
 * An iterator that traverses (but does not modify) a KDTree.
 * This implementation of const_iterator is complete and you
 * will not need to modify it.  However, you may need to use
 * this interface in your solution.
 */
class KDTree::const_iterator : public iterator<forward_iterator_tag, const value_type>
{
public:
    /* Default constructor creates an invalid iterator. */
    const_iterator();

    /* Compare iterators for equality. */
    bool operator== (const_iterator other) const;
    bool operator!= (const_iterator other) const;

    /* Advance to the next element. */
    const_iterator& operator++ ();
    const const_iterator operator++ (int);

    /* Query the current element. */
    const value_type& operator*  () const;
    const value_type* operator-> () const;

private:
    friend class KDTree;

    /* The implementation of this iterator works by maintaining a stack
     * of nodes that it has yet to explore along with the node currently
     * being visited.  This information is represented here.
     */
    deque<const KDTreeNode*> toExplore;
    const KDTreeNode* curr;

    /* Private Constructor: const_iterator(const KDTreeNode* where)
     * Usage: KDTreeNode* node = ...; return const_iterator(node);
     * ----------------------------------------------------
     * Creates a const_iterator that traverses the KDTree starting
     * at the tree node where.  If where is NULL, the iterator acts
     * like an end() iterator.  When implementing KDTree functions
     * that return const_iterators, you will most likely implement
     * those functions by working with raw KDTreeNode*s, then using
     * this constructor to build an iterator to that node.
     */
    const_iterator(const KDTreeNode* where);
};

/******************* Implementation Detail Beyond This Point ************************/

#include <algorithm>

/* Implementation of the KDTree constructor.  This constructor takes in the
 * input range, copies it into a vector, and then recursively builds a tree
 * out of it.  Because this is a template function, we are required to implement
 * it in the .h file.  You will not need to modify this code; however, it may
 * be somewhat edifying to read over.
 */
template <typename InputIterator> KDTree::KDTree(InputIterator begin, InputIterator end)
{
    /* If there isn't any data, then fail gracefully by saying that the tree has
     * dimension zero and that the root is NULL.
     */
    if (begin == end)
    {
        numDimensions = 0;
        numElems = 0;
        root = NULL;
        return;
    }

    /* Construct a vector from the input range.  We accomplish this using a special
     * vector constructor that takes in an iterator range and deep-copies it into
     * the vector.  This is similar to writing
     *
     * vector<value_type> data;
     * copy(begin, end, back_inserter(data));
     *
     * Except that it's much more efficient.
     */
    vector<value_type> data(begin, end);

    /* Build a vector with pointers into the data vector.  We use pointers here
     * instead of actual values because the constructor will need to rearrange
     * vector elements and it is substantially more efficient to shuffle around
     * pointers than to shuffle around actual values.
     */
    vector<value_type*> dataPtrs(data.size());
    for (size_t i = 0; i < dataPtrs.size(); ++i)
        dataPtrs[i] = &data[i];

    /* We're mimicking a map, so keys must be unique.  To ensure this, we
     * sort the data and call the unique algorithm to remove duplicates.  Since
     * elements are key/value pairs, the default sort behavior will sort first
     * by suit, then by value.  However, for unique to work correctly, we'll
     * need to use a custom callback.
     */
    sort(dataPtrs.begin(), dataPtrs.end(), CompareValueTypesIndirectly);
    vector<value_type*>::iterator newEnd =
        unique(dataPtrs.begin(), dataPtrs.end(), ValuesHaveEqualPoints);

    /* The number of elements needs to factor in the number of duplicates. */
    numElems = size_t(distance(dataPtrs.begin(), newEnd));

    /* Obtain the number of dimensions from the first vector. */
    numDimensions = data[0].first.size();

    /* Recursively build the tree. */
    root = recMakeKDTree(dataPtrs.begin(), newEnd, 0, NULL);
}

#endif
